-- DEMOCOM/PRIMA '95
-- Hilfs-FPGA fuer den Democom im Altera 81188                                 C.K.
--------------------------------------------------------------------------------------------------------
-- Dies ist der Ansatz, der sich bis zum Schluss herauskristallisiert hat:
--    - umfassende Befehle vom PC moeglich (Breakpoints ausgeklammert)
--    - Auswahl der Display-Ansteuerung (letztes gueltiges Paar Adresse/Datum oder benutzerdefinierte Zeichen)
--    - Tastatur-Abfrage vom Democom aus moeglich; Hilfs-FPGA wird nur vom PC gesteuert
--    - aktueller Logik-Bedarf: 769 Logikzellen
-- 
--
-- Versionsgeschichte:
-- V0.1   06.04.95
-- V0.15  18.04.95  Zum Ausdrucken
-- V0.16  19.04.95  00.04 Uhr
-- V0.19  19.04.95  21.35 Uhr - morgen grosse Fehlersuche!
-- V0.19b 26.04.95  hackmode, address_inkr und areg_rollin_count hingen von m_state ab, das fuehrt beim dritten eindeutig
--                  zu einem Fehler! Dies ist jetzt die "fehlerbereinigste" Version.
-- V0.19c 28.04.95  Alle Texte aus EPROM, disp_d dauerhaft dort angeschlossen.
-- V0.2a  01.05.95  Hoffentlich alle Fertigkeiten abgedeckt, Beginn der Simulation
-- V0.2b  05.05.95  Simuliert; scheint zu tun. Einbau von "Anregungen" der WiMis
-- V0.22  10.05.95  Instanz des pc_if eingebaut. Groesse in Altera-FPGAs: 1063 LCs (411 fuer pc_if)
-- V0.23  11.05.95  Design wurde zu ctrl_pur: "alle" Spuren der Display-Ansteuerung entfernt
-- V0.23 bis V0.9   Zahllose Spezifikationserweiterungen und Fehlerbereinigungen
-- V0.9   23.06.95  In Hardware lauffaehige Debugging-Version.
--                  Enthaelt noch die Signale debugclk sowie debugstate und debug_readwrite.
--                  Display-Ansteuerung, bei Benutzung durch Democom? Tastaturabfrage?
-- V1.0   19.10.95  Test dieser Version (mit User Mode beim Display, Tastaturabfrage)
-- V1.01  01.11.95  a/dreg_muxout werden aktualisiert, wenn (a_or_d_changed = '1'), also auch, wenn das Hilfs-FPGA
--                  selbst schreibt/liest.
-- V1.1   15.11.95  Alle Features funktionieren zuverlaessig (eingeschlossen Konfiguration, Tastatur, Display)
-- V1.3   22.01.96  Ueberarbeitete Kommentare
-- V1.31  29.05.96  Fehler bei Lesen aus Adressbereich von Display/Tastatur beseitigt (address wurde getrieben)
-- V1.32  30.05.96  Erst so ist das Lesen aus Adressbereich von Display/Tastatur korrekt (abhaengig von mem_read)
--------------------------------------------------------------------------------------------------------

-- Kleine Terminologie:
--
-- Democom: die CPU auf dem Mikroprozessorboard, residierend in dem 81500-FPGA
-- 81500: das FPGA vom Typ Altera FLEX8000 81500 fuer die CPU des Mikroprozessorboards
-- Hilfs-FPGA: der "Chipsatz" auf dem Mikroprozessorboard, residierend in dem 81188-FPGA
-- 81188: das FPGA vom Typ Altera FLEX8000 81188 fuer das Hilfs-FPGA
-- Output Enable (OE): Tristate-Treiber besitzen stets einen solchen Anschluss. Wird er aktiviert,
--                     wird der Ausgang des Treibers mit dem Wert des Einganges getrieben. Ansonsten ist
--                     der Ausgang "offen" (hochohmig). Auch Speicherbausteine besitzen fuer ihre Datenleitungen
--                     einen OE-Anschluss.
-- 8255: Der Portbaustein von Intel, der den bidirektionalen Bus zwischen PC und Hilfs-FPGA ermoeglicht
--       (acht Bit Daten, vier Bit Kontrollsignale).
-- Nibble: Ein halbes Byte (vier Bits)
-- FSM: finite state machine, ein endlicher Automat
-- LSB: least significant bit,  niedrigstwertiges Bit


library ieee;
use ieee.std_logic_1164.all;

entity ctrl_pur is
    port(
        --
        -- Aufgrund einer Unvertraeglichkeit mit Maxplus2 sind keine Signal-/Variablenname mit fuehrendem "n_" erlaubt.
        --
        -- Deswegen gilt: Signale, deren Namen mit "m_" oder "progr_m_" beginnen, sind low-aktiv, alle anderen high-aktiv.
        --
        -- Ausnahmen: "reset_unsync", "chipsel"-Signale, "reset_democom"
        --
        -- Beachte: Wurden endliche Automaten beschrieben, so heissen die Signale fuer gegenwaertigen und
        -- naechsten Zustand z.B. c_state und m_state (i.e. current_state und next_state).
        --
        
        --- Eingaenge ---
        reset_unsync:   in std_logic;                        -- KEIN Power-On-Reset, sondern zur Initialisierung 
                                                             -- nach der Konfiguration oder spaeter
                                                             -- Ist noch unsynchronisiert mit der Clock - das in allen
                                                             -- (Sub)Designs verwendete Signal "reset" ist dieses hier,
                                                             -- ueber ein Flipflop synchronisiert.
                                                             -- Ist low-aktiv.
                                                             
        clk:            in std_logic;                        -- globale Clock fuer alle speichernden Elemente
        clk_invers:     in std_logic;
        
        progr_m_status: in std_logic;                        -- Rueckmeldung bei Konfiguration des Democom:
                                                             -- kurz '0' zu Beginn; waehrend der Konf. '0', falls Fehler

        m_astrobe:      in std_logic;                        -- Meldung vom Democom: gueltige Adresse liegt an.
                                                             -- Wird nicht benutzt, stattdessen wird a_or_d_changed
                                                             -- gebildet.
        
        m_reg_ack:      in std_logic;                        -- Bestaetigung vom Democom: Register Acknowledge -  
                                                             -- Wert eines internen Registers liegt auf Adress-/Datenbus
                                                             -- Um keine zusaetzlichen Datenpfade in einem Democom-Design
                                                             -- zu benoetigen und da interne Register breiter als 16 Bit
                                                             -- sein koennen, braucht der mit m_reg_req (Register Request)
                                                             -- angeforderte Wert eines Registers nur auf Adress- ODER
                                                             -- Datenbus zu liegen; beide werden gepuffert und am PC
                                                             -- angezeigt, wo der "richtige" Wert ausgewaehlt werden kann. 
                                                             
        m_halt_ack:     in std_logic;                        -- Bestaetigung vom Democom: habe angehalten
        
        progr_conf_done: in std_logic;                       -- Bestaetigung vom 81500: Konfiguration beendet
                                                             -- (d.h. der 81500-Baustein aus Alteras FLEX8000-Serie
                                                             -- wurde von einer "tabula rasa", die er nach jeder
                                                             -- Unterbrechung der Versorgungsspannung wieder ist,  zu
                                                             -- z.B. dem Democom oder dem Ladedesign
                                                             -- fuer das Mikrocode-RAM gemacht)
        
        m_obf_unsync:   in std_logic;                        -- Das Hilfs-FPGA kommuniziert ueber ein acht Bit breites
                                                             -- Interface mit einem externen PC. Da diese Leitungen
                                                             -- bidirektional genutzt werden, wird ein 
                                                             -- Handshaking-Protokoll benoetigt. Der verwendete 
                                                             -- Portbaustein 8255 von Intel stellt dieses auf vier
                                                             -- Leitungen zur Verfuegung:
                                                             --   - m_obf='0' -> PC hat Daten an Hilfs-FPGA gesandt
                                                             --   - m_ack='0' -> Hilfs-FPGA holt Daten vom 8255 ab
                                                             --   - ibf='0'   -> Hilfs-FPGA hat Daten an PC gesandt
                                                             --   - m_stb='0' -> Hilfs-FPGA sendet Daten zum 8255 
                                                             --
                                                             -- "Output buffer full" am 8255, d.h. ein Befehl oder Daten
                                                             -- vom PC liegen im 8255 zur Abholung bereit.
                                                             -- Wird vor Verwendung noch mit der Clock synchronisiert.
                                                             
        ibf_unsync:     in std_logic;                        -- "Input buffer full" am 8255, d.h. Hilfs-FPGA hat Daten
                                                             -- zum 8255 geschrieben, die noch nicht abgeholt worden sind.
                                                             -- Wird vor Verwendung noch mit der Clock synchronisiert.

        tast_reihe:     in std_logic_vector(1 to 4);         -- Die "Hex-Tastatur" (20 T.) auf der Mikrorechner-Platine
                                                             -- ist in vier Reihen zu je fuenf Tasten (Schalter) 
                                                             -- organisiert, wobei die Ausgaenge der Schalter einer 
                                                             -- Reihe alle das selbe Signal tast_reihe(x) bedienen.
                                                             -- Die Tasten einer Reihe sind fuer das Hilfs-FPGA dennoch
                                                             -- unterscheidbar, da sie in unterschiedlichen Spalten
                                                             -- liegen, von denen genau eine mit '1' getrieben wird.
                                                             -- Kurzschluesse bei Druecken mehrerer Tasten werden 
                                                             -- ueber Schutzdioden verhindert.
                                                             
        disp_d7_in:     in std_logic;                        -- achtes Datenbit des Displays mit 20x2 Zeichen -
                                                             -- ist bidirektional und wird entweder vom Hilfs-FPGA oder
                                                             -- dem Display getrieben; musste mit einem externen
                                                             -- Tristate-Treiber realisiert werden.
                                                             -- eingehende Richtung: BusyFlag (Display beschaeftigt)
                                                             -- ausgehende Richtung: normales Datenbit
                                                             
        --- Bidirektionale Signale ---
        databus:        inout std_logic_vector(15 downto 0); -- Datenbus der Mikrorechner-Platine
        address:        inout std_logic_vector(19 downto 0); -- Adressbus der Mikrorechner-Platine
        
        pc_if_data:     inout std_logic_vector(7 downto 0);  -- PC-Interface: Daten zwischen Hilfs-FPGA und PC
        
        m_oe:           inout std_logic;                     -- "Output Enable" an I/O, d.h. durch Adresse selektiertes
                                                             -- Geraet (SRAM, EPROM, Par.I/O, Ser.I/O) treibt Datenbus
                                                             
        m_we:           inout std_logic;                     -- "Write Enable" an I/O, d.h. durch Adresse selektiertes
                                                             -- Geraet (SRAM, EPROM, Par.I/O, Ser.I/O) liest von Datenbus
        
        --- Ausgaenge ---
        m_ack:          out std_logic;                       -- "Acknowledge" am 8255, d.h. Hilfs-FPGA hat via (m_obf='0')
                                                             -- erkannt, dass Daten im Portbaustein bereit liegen und
                                                             -- laesst diesen auf pc_if_data treiben.
                                                             
        m_stb:          out std_logic;                       -- "Strobe" am 8255, d.h. der Wert von pc_if_data wird im
                                                             -- Portbaustein gepuffert (worauf ibf auf '1' geht).
                                                             
                                                             -- Fuer Display 20x2 Zeichen:
        disp_d:         out std_logic_vector(6 downto 0);    --   untere sieben Datenbits (achtes ist bidirektional)
        disp_e:         out std_logic;                       --   Enable: Zeichen ausgeben, Cursor wird weitergerueckt
        disp_rw:        out std_logic;                       --   Read/Write - '1' zum Abfragen des BusyFlag
        disp_rs:        out std_logic;                       --   '1': "normales" ASCII-Zeichen, '0': Steuerzeichen

        clk1_democom:   out std_logic;                       -- globale Clock fuer Democom - ergibt sich aus
                                                             -- (clk AND clk_aktiv), d.h. im Regelfalle laeuft der Democom
                                                             -- mit demselben Takt wie das Hilfs-FPGA, dieser kann ihm
                                                             -- aber, im Single Step-Modus naemlich, entzogen werden. 
        clk2_democom:   out std_logic;                       -- invertierte globale Clock fuer Democom

        reset_democom:  out std_logic;                       -- Reset fuer den Democom - 
                                                             -- KEIN Power-On-Reset, sondern zur Initialisierung 
                                                             -- nach der Konfiguration oder spaeter
                                                             -- Ist low-aktiv.
                                                             
        m_highz:        out std_logic;                       -- Democom soll vom Bus gehen, Hilfs-FPGA uebernimmt
        
                                                             -------------------------------------------------------------
                                                             -- In Abhaengigkeit von der Adresse soll hoechstens ein
                                                             -- Ein-/Ausgabe-Geraet aktiviert sein.
                                                             -- Kein Geraet darf aktiv sein, falls der Adressbus nicht
                                                             -- getrieben wird (wg Pullups erscheint die Adresse $FFFFF),
                                                             -- eine nicht vergebene Adresse anliegt oder das Mikrocode-
                                                             -- RAM adressiert wird (hier intern vorhandenes 21. Adressbit
                                                             -- ist '1').
                                                             -- Ansonsten ist genau eines der folgenden low-aktiven
                                                             -- Signale '0':
        chipsel_sram:   out std_logic;                       -- low-aktives Chipselect-Signal SRAM
        chipsel_pio:    out std_logic;                       -- low-aktives Chipselect-Signal parallele I/O
        chipsel_sio:    out std_logic;                       -- low-aktives Chipselect-Signal serielle I/O
        chipsel_rom:    out std_logic;                       -- low-aktives Chipselect-Signal optionales ROM
                                                             -------------------------------------------------------------
        
        progr_m_config: out std_logic;                       -- Beginn der Konfiguration (Programmierung) des 81500
        progr_dclk:     out std_logic;                       -- Clock fuer Konfiguration des 81500
        progr_data:     out std_logic;                       -- 1 Bit Daten fuer (serielle) Konfiguration des 81500
         
        m_reg_req:      out std_logic;                       -- Aufforderung an Democom: Register Request -
                                                             -- Das durch reg_select bezeichnete Register auf
                                                             -- Adress- ODER Datenbus legen.
                                                             
        reg_select:     out std_logic_vector(3 downto 0);    -- 16 verschiedene interne (Register-)Werte eines Democom
                                                             -- koennen ausgegeben werden; Kodierung nicht vorgegeben.
                                                             -- Auf dieses Signal nur achten, wenn m_reg_req '0' ist.
                                                             
        m_halt:         out std_logic;                       -- Aufforderung an Democom: Anhalten (sinnvollerweise bei
                                                             -- naechstem Fetch/Decode) - Bestaetigung mit m_halt_ack.

        tast_spalte:    out std_logic_vector(1 to 5);        -- siehe tast_reihe; genau ein Signal ist '1', der Rest '0'.
        
        disp_d7_out:    out std_logic;                       -- siehe disp_d7_out
        disp_d7_oe:     out std_logic );                     -- low-aktives Output Enable fuer disp_d7_out
        
end ctrl_pur;


architecture behaviour of ctrl_pur is                        -- Dies ist eine Verhaltensbeschreibung.

component display
--
-- Primitive Ansteuerung der 20x2-Zeichen-Anzeige.
-- Angezeigt wird entweder: [Letzte gueltige Adresse in fuenf Hex-Ziffern] [Zeilenvorschub]
--                          [Letztes gueltiges Datum in vier Hex-Ziffern]
-- Oder: [beliebige Zeichen vom Democom]
--
    port(
        reset:          in std_logic;                        -- KEIN Power-On-Reset, sondern zur Initialisierung 
                                                             -- nach der Konfiguration oder spaeter
                                                             -- Ist low-aktiv und mit der globalen Clock synchronisiert.
                                                             
        clk:            in std_logic;                        -- globale Clock fuer alle speichernden Elemente
        clk_invers:     in std_logic;
        
        a_or_d_changed: in std_logic;                        -- Benachrichtigung von ctrl_pur: Adresse oder Datum
                                                             -- waren gueltig und wurden gepuffert; nun kann dieses
                                                             -- Subdesign sie in aller Ruhe anzeigen.
                                                             
        areg_muxout:    in std_logic_vector(19 downto 0);    -- Puffert bei (a_or_d_changed = '1') die Adresse
        dreg_muxout:    in std_logic_vector(15 downto 0);    -- Puffert bei (a_or_d_changed = '1') die Daten
        
        user_char:      in std_logic_vector(8 downto 0);     -- Statt der letzten gueltigen Werte von Adress- und
                                                             -- Datenbus koennen nun auch beliebige Texte vom
                                                             -- Democom auf dem Display ausgegeben werden.
                                                             -- user_char(7 downto 0) enthaelt das gerade anzu-
                                                             -- zeigende Zeichen, user_char(8) den dazugehoerigen
                                                             -- Wert von disp_rs.

        user_mode_set:  in std_logic;                        -- | Mit diesen beiden Signalen wird das Flipflop
        user_mode_reset:in std_logic;                        -- | "user_mode_ff" gesetzt/geloescht; ist dieses gesetzt,
                                                             -- | so wird user_char angezeigt.
        
        user_disp_go:   in std_logic;                        -- '1': Anzeige von user_char beginnen
        
        disp_d7_in:     in std_logic;                        -- Display - BusyFlag auf achter Datenleitung
        disp_rw:        out std_logic;                       -- Display - Read/Write
        disp_e:         out std_logic;                       -- Display - Enable (Data Strobe)
        disp_rs:        out std_logic;                       -- Display - 1: "normales" ASCII-Zeichen
        disp_d:         out std_logic_vector(6 downto 0);    -- Display - Daten (hier 7 Bits, achtes ist bidirektional)
        
        user_disp_busy: out std_logic;                       -- '1': Zeichen vom Democom ist noch nicht angezeigt worden
        
        disp_d7_out:    out std_logic;                       -- Display - achtes Datenbit auf achter Datenleitung
        disp_d7_oe:     out std_logic );                     -- Display - Output Enable fuer disp_d7_out
end component;
    
component pc_if
    port(
        reset:          in std_logic;                        -- KEIN Power-On-Reset, sondern zur Initialisierung 
                                                             -- nach der Konfiguration oder spaeter
                                                             
        clk:            in std_logic;                        -- globale Clock fuer alle speichernden Elemente
        clk_invers:     in std_logic;
        
        progr_m_status: in std_logic;                        -- Rueckmeldung bei Konfiguration des 81500:
                                                             -- kurz '0' zu Beginn; waehrend der Konf. '0', falls Fehler
                                                             
        progr_conf_done: in std_logic;                       -- Bestaetigung vom 81500: Konfiguration beendet
        
        m_obf:          in std_logic;                        -- "Output buffer full" am 8255, d.h. ein Befehl oder Daten
                                                             -- vom PC liegen im 8255 zur Abholung bereit
                                                             
        ibf:            in std_logic;                        -- "Input buffer full" am 8255, d.h. Hilfs-FPGA hat Daten
                                                             -- zum 8255 geschrieben, die noch nicht abgeholt worden sind.
                                                             
        mem_ready:      in std_logic;                        -- Bestaetigung von RAM-Zugriffsmaschine (in dieser Datei):
                                                             -- Schreib-/Lesevorgang abgeschlossen
        
        m_reg_ack:      in std_logic;                        -- Bestaetigung vom Democom:  gewuenschtes Register liegt an
        
        areg_muxout:    in std_logic_vector(19 downto 0);    -- Puffert bei (a_or_d_changed = '1') die Adresse
        dreg_muxout:    in std_logic_vector(15 downto 0);    -- Puffert bei (a_or_d_changed = '1') die Daten
        
        m_halt_ack:     in std_logic;                        -- Bestaetigung vom Democom: habe angehalten
        
        pc_if_data_in:  in std_logic_vector(7 downto 0);     -- Daten zwischen Hilfs-FPGA und PC - eingehende Richtung
        pc_if_data_out: out std_logic_vector(7 downto 0);    -- Daten zwischen Hilfs-FPGA und PC - ausgehende Richtung
        
        mem_read:       out std_logic;                       -- |Aufforderung an RAM-Zugriffsmaschine (in dieser Datei):
        mem_write:      out std_logic;                       -- |eine Speicherzelle beschreiben/auslesen
        
        m_ack:          out std_logic;                       -- "Acknowledge" am 8255, d.h. 8255 treibt Datenbus
        m_stb:          out std_logic;                       -- "Strobe" am 8255, d.h. Daten werden in 8255 uebernommen
        
        progr_m_config: out std_logic;                       -- Beginn der Konfiguration des Democom
        progr_dclk:     out std_logic;                       -- Clock fuer Konfiguration des Democom
        
        m_reg_req:      out std_logic;                       -- |Aufforderung an Democom: das durch reg_select
        reg_select:     out std_logic_vector(3 downto 0);    -- |bezeichnete Register auf Adress- ODER Datenbus legen
        
        arollin:        out std_logic;                       -- | 20 bzw. 16 Bit breite Register [a/d]reg_rollin 
        drollin:        out std_logic;                       -- | werden mit Daten vom PC-Interface bei ([a/d]rollin='1') 
        arollin_mux:    out std_logic;                       -- | und Rueckflanke vierbitweise gefuellt; [a/d]rollinmux 
        drollin_mux:    out std_logic;                       -- | waehlt ein Nibble von pc_if_data aus.
        
        pc_if_driven:   out std_logic;                       -- Output Enable fuer pc_if_data_out
        
        clk_aktiv:      out std_logic;                       -- Democom-Clock <= (clk AND clk_aktiv);
        
        m_halt:         out std_logic;                       -- Aufforderung an Democom: vor naechstem Befehl anhalten
        
        progr_data:     out std_logic;                       -- 1 Bit Daten fuer (serielle) Konfiguration des Democom
         
        reset_democom:  out std_logic );                     -- Reset, der an den Democom geht - 
                                                             -- KEIN Power-On-Reset, sondern zur Initialisierung 
                                                             -- nach Einschalten und Konfiguration des Democom
end component;

signal chipsel: std_logic_vector(5 downto 0);       -- (4 downto 0) werden auf chipsel_sram, -pio, -sio, -rom zugewiesen;
                                                    -- chipsel(0) zeigt Speicherzugriffe fuer den User Mode des Displays
                                                    -- an, chipsel(5) fuer die Hex-Tastatur (20 Tasten).

signal areg_muxout:  std_logic_vector(19 downto 0); -- Puffert bei (a_or_d_changed = '1') den Wert des Adressbusses
                                                    -- zur byteweisen Uebertragung an den PC

signal areg_rollin:  std_logic_vector(20 downto 0); -- Gegenstueck zu areg_muxout; Bytes vom PC werden gesammelt,
                                                    -- um auf einmal die 20 Bits des Adressbusses treiben zu koennen.
                                                    -- Fuer die Umsetzung der Bitbreiten waere ein Hineinschieben von je
                                                    -- acht Bits in dieses Register am simpelsten, es werden jedoch pro
                                                    -- Byte zweimal vier Bit eingeschoben, um die Moeglichkeit einer
                                                    -- Hex-Tastatur als zweite Datenquelle zu erhalten.
                                                    --
                                                    -- 21. Bit: '0' - "normale" Adresse (SRAM, I/O, ROM, Display, Tast.)
                                                    --          '1' - Adresse im Mikrocode-RAM

signal dreg_muxout:  std_logic_vector(15 downto 0); -- Puffert bei (a_or_d_changed = '1') den Wert des Datenbusses
                                                    -- zur byteweisen Uebertragung an den PC.
                                                    
signal dreg_rollin:  std_logic_vector(15 downto 0); -- Gegenstueck zu dreg_muxout; Bytes vom PC werden gesammelt,
                                                    -- um auf einmal die 16 Bits des Datenbusses treiben zu koennen.
                                                    -- Weiteres siehe areg_rollin.
 
signal arollin, drollin: std_logic;                 -- Steuersignale: Daten in areg-/dreg_rollin hinein "shiften"

signal arollin_mux, drollin_mux: std_logic;         -- Obere oder untere vier Bit der Daten vom PC hineinschieben.

signal mem_read, mem_write: std_logic;              -- Steuersignale an RAM-Interface: Mit Lese-/Schreibzyklus beginnen.
                                                    -- mem_read/mem_write bleiben bis zu dessen Ende '1'!
                                                                                                     
signal mem_ready: std_logic;                        -- Rueckmeldung vom RAM-Interface: Daten geschrieben/gelesen.
                                                    
signal m_highz_t: std_logic;                        -- Identisch mit m_highz, kann aber ausgelesen werden.

signal clk_aktiv: std_logic;                        -- Democom-Clock <= (clk AND clk_aktiv);

signal a_or_d_changed: std_logic;

signal pc_if_data_in:  std_logic_vector(7 downto 0);    
signal pc_if_data_out: std_logic_vector(7 downto 0);    

signal pc_if_driven: std_logic;       -- Daten fuer das PC-Interface, werden ueber Tristatebuffer nach                   
                                      -- aussen gegeben, falls (pc_if_driven='1')
                                      
signal reset: std_logic;

signal m_obf: std_logic;
signal ibf:   std_logic;

signal user_char: std_logic_vector(8 downto 0);
signal user_mode_set, user_mode_reset, user_disp_go, user_disp_busy: std_logic;

signal tast_reg:        std_logic_vector(19 downto 0);  -- Speichert, welche Tasten gedrueckt wurden, wird geloescht
                                                        -- durch Auslesen.
signal tast_reg_clear:  std_logic_vector(1 downto 0);
signal tast_spalte_t:   std_logic_vector(1 to 5);       -- Identisch mit tast_spalte, aber beschreibbar

type ram_state_type is (warten, read_init, read_ready, write_init, write_ready);                 -- 5 Zustaende
signal c_rasta, m_rasta: ram_state_type; -- current_state und next_state der RAM-Zugriffsmaschine
                                                   

begin
    pc_interface: pc_if port map(reset, clk, clk_invers, 
                                 progr_m_status, progr_conf_done, m_obf, ibf, mem_ready, m_reg_ack, 
                                 areg_muxout, dreg_muxout, m_halt_ack, pc_if_data_in, pc_if_data_out, 
                                 mem_read, mem_write, m_ack, m_stb,
                                 progr_m_config, progr_dclk, m_reg_req, reg_select,
                                 arollin, drollin, arollin_mux, drollin_mux, pc_if_driven,
                                 clk_aktiv, m_halt, progr_data, reset_democom);
                                 
    display_control: display port map(reset, clk, clk_invers, 
                                      a_or_d_changed, areg_muxout, dreg_muxout, 
                                      user_char, user_mode_set, user_mode_reset, user_disp_go, 
                                      disp_d7_in, disp_rw, disp_e, disp_rs, disp_d,
                                      user_disp_busy, disp_d7_out, disp_d7_oe);


--- Display Anfang ---
    user_mode_set <= (NOT chipsel(0)) AND address(1) AND databus(0) AND m_oe AND (NOT m_we); 
                                                                               -- Beschreiben einer Adresse auf "1-" im
                                                                               -- Adressbereich des Displays mit "...1"
                                                                               
    user_mode_reset <= (NOT chipsel(0)) AND address(1) AND (NOT databus(0)) AND m_oe AND (NOT m_we); 
                                                                               -- Beschreiben einer Adresse auf "1-" im
                                                                               -- Adressbereich des Displays mit "...0"
    
    user_disp_go <= (NOT chipsel(0)) AND (NOT address(1)) AND (NOT address(0)) AND m_oe AND (NOT m_we); 
                                                                               -- Beschreiben einer Adresse auf "00" im
                                                                               -- Adressbereich des Displays
    USER_DISPLAY_CHAR: process(reset, clk, chipsel, address, databus, m_oe, m_we)
    begin
      if (reset = '0') then
        user_char <= "000000000";
      elsif (clk'event and clk='1') then  -- Vorderflanke
        if (chipsel(0) & address(1 downto 0) & m_oe & m_we = "00110") then                                                                              
                                                                               -- Beschreiben einer Adresse auf "01" im
                                                                               -- Adressbereich des Displays
          user_char <= databus(8 downto 0);
        else
          user_char <= user_char;
        end if;
      end if;
    end process;
--- Display Ende ---
    
    
--- Tastatur Anfang ---
    -- So war es einmal gedacht, um Kurzschluesse bei Druecken mehrerer Tasten
    -- zu vermeiden (deswegen der Name tast_spalte_z).
    -- Aufgegeben wegen Mangels an Output Enable-Signalen auf 81188 (nur vier),
    -- stattdessen fuenf Schutzdioden.
    -- tast_spalte_z(x) <= '1' when (tast_spalte_t(x) = '0') else 'Z';  1 <= x <= 5
    
    tast_spalte <= tast_spalte_t;
    
    TASTATUR_ANSTEUERUNG: process(reset, clk_invers, tast_spalte_t)
    begin
      if (reset = '0') then
        tast_spalte_t <= "01111";
      elsif(clk_invers'event and clk_invers='1') then  -- Rueckflanke
        case tast_spalte_t is
          when "10000" => tast_spalte_t <= "01000";  -- | genau eine Spalte ist aktiviert, die Spalten werden
          when "01000" => tast_spalte_t <= "00100";  -- | reihum aktiviert
          when "00100" => tast_spalte_t <= "00010";
          when "00010" => tast_spalte_t <= "00001";
          when "00001" => tast_spalte_t <= "10000";
          when others  => tast_spalte_t <= "10000";
        end case;
      end if;
    end process;
    
    TASTATUR_REGISTER: process(reset, clk, tast_reihe, tast_spalte_t, chipsel, address, m_oe, m_we)
    --
    -- Von oben nach unten enthaelt die IF-Anweisung folgende Faelle:
    --
    --   - Zuruecksetzen bei Reset
    --
    --   - Lesen aus dem Adressbereich der Tastatur -> ein Teil von tast_reg wird ausgegeben (konkurrent hierzu
    --     im Prozess BUSSE), je nachdem, ob eine gerade oder ungerade Adresse vorliegt. Dieser Teil wird durch
    --     Setzen eines Bits von tast_reg_clear zum Loeschen vorgesehen.
    --
    --   - Die untere Haelfte von tast_reg ist zum Loeschen vorgesehen - sobald das aktuelle Lesen beendet wurde,
    --     wird diese ebenso wie das korrespondierende Bit von tast_reg_clear geloescht. 
    --
    --   - ebenso fuer die obere Haelfte von tast_reg. Nicht mit obigem Fall zusammenfassen!
    --
    --   - Setzen von Bits in tast_reg, d.h. Detektieren von Tastendruecken. Ein Bit wird '1', wenn entweder eine
    --     eindeutige Kombination aus Reihe und Spalte jeweils '1' ist, oder wenn es bereits '1' war (Bits werden 
    --     nur durch Auslesen geloescht).
    --   
    begin
      if (reset = '0') then
        tast_reg       <= "00000000000000000000";
        tast_reg_clear <= "00";
        
      elsif(clk'event and clk='1') then  -- Vorderflanke
      
        if ((chipsel(5)='0') AND (m_oe='0') AND (m_we = '1')) then         -- Lesen aus dem Adressbereich der Tastatur
          if (address(0) = '0') then                       -- gerade Adresse -> tast_reg(15..0) ausgeben, dann loeschen
            tast_reg                   <= tast_reg;
            tast_reg_clear(1 downto 0) <= tast_reg_clear(1) & "1";
          else                                             -- ungerade Adresse -> tast_reg(19..16) ausg., dann loeschen
            tast_reg                   <= tast_reg;
            tast_reg_clear(1 downto 0) <= "1" & tast_reg_clear(0);
          end if;
          
        elsif(tast_reg_clear(0) = '1') then                -- untere Haelfte von tast_reg zum Loeschen vorgesehen
          if (m_oe = '1') then                             -- Lesen beendet, nun kann geloescht werden
            tast_reg(19 downto 16)     <= tast_reg(19 downto 16);
            tast_reg(15 downto  0)     <= "0000000000000000";
            tast_reg_clear(1 downto 0) <= tast_reg_clear(1) & "0";
          else
            tast_reg                   <= tast_reg;
            tast_reg_clear             <= tast_reg_clear;
          end if;
          
        elsif(tast_reg_clear(1) = '1') then
          if (m_oe = '1') then                             -- obere Haelfte von tast_reg zum Loeschen vorgesehen
            tast_reg(19 downto 16)     <= "0000";          -- Lesen beendet, nun kann geloescht werden
            tast_reg(15 downto 0)      <= tast_reg(15 downto 0);
            tast_reg_clear(1 downto 0) <= "0" & tast_reg_clear(0);
          else
            tast_reg                   <= tast_reg;
            tast_reg_clear             <= tast_reg_clear;
          end if;
          
        else
          tast_reg(0)  <= (tast_reihe(1) AND tast_spalte_t(1)) OR tast_reg(0); -- | War tast_reg(x) schon '1', so bleibt es
          tast_reg(1)  <= (tast_reihe(1) AND tast_spalte_t(2)) OR tast_reg(1); -- | dieses; sonst wird es genau dann '1',
          tast_reg(2)  <= (tast_reihe(1) AND tast_spalte_t(3)) OR tast_reg(2); -- | falls "seine" Spalte aktiv ('1') war
          tast_reg(3)  <= (tast_reihe(1) AND tast_spalte_t(4)) OR tast_reg(3); -- | und "seine" Zeile '1' meldet.
          tast_reg(4)  <= (tast_reihe(1) AND tast_spalte_t(5)) OR tast_reg(4);
          tast_reg(5)  <= (tast_reihe(2) AND tast_spalte_t(1)) OR tast_reg(5);
          tast_reg(6)  <= (tast_reihe(2) AND tast_spalte_t(2)) OR tast_reg(6);
          tast_reg(7)  <= (tast_reihe(2) AND tast_spalte_t(3)) OR tast_reg(7);
          tast_reg(8)  <= (tast_reihe(2) AND tast_spalte_t(4)) OR tast_reg(8);
          tast_reg(9)  <= (tast_reihe(2) AND tast_spalte_t(5)) OR tast_reg(9);
          tast_reg(10) <= (tast_reihe(3) AND tast_spalte_t(1)) OR tast_reg(10);
          tast_reg(11) <= (tast_reihe(3) AND tast_spalte_t(2)) OR tast_reg(11);
          tast_reg(12) <= (tast_reihe(3) AND tast_spalte_t(3)) OR tast_reg(12);
          tast_reg(13) <= (tast_reihe(3) AND tast_spalte_t(4)) OR tast_reg(13);
          tast_reg(14) <= (tast_reihe(3) AND tast_spalte_t(5)) OR tast_reg(14);
          tast_reg(15) <= (tast_reihe(4) AND tast_spalte_t(1)) OR tast_reg(15);
          tast_reg(16) <= (tast_reihe(4) AND tast_spalte_t(2)) OR tast_reg(16);
          tast_reg(17) <= (tast_reihe(4) AND tast_spalte_t(3)) OR tast_reg(17);
          tast_reg(18) <= (tast_reihe(4) AND tast_spalte_t(4)) OR tast_reg(18);
          tast_reg(19) <= (tast_reihe(4) AND tast_spalte_t(5)) OR tast_reg(19);
          tast_reg_clear <= tast_reg_clear;
        end if;
      end if;
    end process;
--- Tastatur Ende ---    
    
    
    a_or_d_changed <= '1' when ((m_oe = '1') AND (m_we = '0')) OR          -- Jemand schreibt in Speicher oder
                               ((m_oe = '0') AND (m_we = '1')) OR          -- jemand liest aus Speicher oder
                               (m_reg_ack = '0')                           -- Democom-Register ist da
                      else '0';
    
    
    pc_if_data    <= pc_if_data_out when (pc_if_driven = '1') else "ZZZZZZZZ"; -- Getrennte In-/Out-Busse wie
                                                                               -- in altera1500.tdf
    pc_if_data_in <= pc_if_data;


    clk1_democom <= clk AND clk_aktiv;        -- Democom bekommt Takt wie Hilfs-FPGA, maskiert mit
    clk2_democom <= NOT (clk AND clk_aktiv);  -- clk_aktiv aus dem PC-Interface.

    
    m_highz <= m_highz_t;                                                    -- | Democom muss Busse freigeben, falls
    m_highz_t <= '0' when ((mem_read = '1') OR (mem_write = '1')) else '1';  -- | Hilfs-FPGA Speicher schreibt/liest
    
    
--- Synchronisation Anfang ---    
    SYNCHRONER_RESET: process(clk_invers, reset_unsync)
    begin
        if (clk_invers'event and clk_invers='1') then  -- Rueckflanke
            if (reset_unsync = '0') then
                reset <= '0';
            else
                reset <= '1';
            end if;
        end if;
    end process;
    
    SYNCHRONES_M_OBF: process(clk_invers, m_obf_unsync)
    begin
        if (clk_invers'event and clk_invers='1') then  -- Rueckflanke
            if (m_obf_unsync = '0') then
                m_obf <= '0';
            else
                m_obf <= '1';
            end if;
        end if;
    end process;
    
    SYNCHRONES_IBF: process(clk_invers, ibf_unsync)
    begin
        if (clk_invers'event and clk_invers='1') then  -- Rueckflanke
            if (ibf_unsync = '1') then
                ibf <= '1';
            else
                ibf <= '0';
            end if;
        end if;
    end process;
--- Synchronisation Ende ---    
    
--- Chipselect Anfang ---    
    -- Tastatur        <= chipsel(5);
    chipsel_sram       <= chipsel(4);
    chipsel_pio        <= chipsel(3);
    chipsel_sio        <= chipsel(2);
    chipsel_rom        <= chipsel(1);
    -- Display         <= chipsel(0);
    
    CHIP_SELECT: process(address, mem_write, mem_read, areg_rollin)
    begin
      if ((mem_write = '0' AND mem_read = '0') OR areg_rollin(20) = '0') then
      --
      -- SRAM/PIO/SIO/ROM koennen Chipselect bekommen, falls entweder das Hilfs-FPGA _nicht_ den Adressbus treibt, oder
      -- es das zwar tut, aber im Bereich ebendieser Komponenten (21. Adressbit ist '0').
      -- Verhindert wird hierdurch, dass sowohl SRAM/PIO/SIO/ROM einen Wert zurueckliefert, als auch das von diesen
      -- unabhaengige Mikrocode-RAM (wenn der 81500 ein sogenanntes Mikrocode-Ladedesign enthaelt).
      --
        case address(19 downto 17) is
          when "000"  => chipsel <= "101111";                  -- 00000-1FFFF -> SRAM aktiv
          when "100"  => case address(16 downto 15) is
                           when "00"   => chipsel <= "110111"; -- 80000-87FFF -> Par. I/O aktiv
                           when "01"   => chipsel <= "111011"; -- 88000-8FFFF -> Ser. I/O aktiv
                           when "10"   => chipsel <= "111110"; -- 90000-97FFF -> Display
                           when "11"   => chipsel <= "011111"; -- 98000-9FFFF -> Tastatur 
                           when others => chipsel <= "111111"; -- keine Komponente aktiv bei ungueltiger Adresse
                         end case;
          when "110"  => chipsel <= "111101";                  -- C0000-DFFFF -> EPROM aktiv (eigtl.: C0000-CFFFF)
          when "001" | "010" | "011" | "101" | "111" 
                      => chipsel <= "111111"; -- keine Komponente aktiv, insbesondere auch dann, wenn der Adressbus nicht
                                              -- getrieben wird ("ZZZ" wird ueber Pullups zu "111")
          when others => chipsel <= "111111"; -- keine Komponente aktiv bei ungueltiger Adresse
        end case;
      else
        chipsel <= "111111"; -- keine Komponente aktiv, da Adresse im Mikrocode-RAM
      end if;
    end process;
--- Chipselect Ende --- 

    BUSSE: process(mem_write, mem_read, areg_rollin, dreg_rollin, chipsel, address, m_oe, m_we, user_disp_busy, tast_reg)
    begin
    -- Busquellen-Multiplexer:
        if (mem_write = '1') then
        -- Das Hilfs-FPGA beschreibt eine Speicherzelle und treibt dazu Adress- und Datenbus.
        -- Der Democom wird automatisch von den Bussen abgeklemmt.
        
          address <= areg_rollin(19 downto 0);
          databus <= dreg_rollin;
        
        
        elsif (mem_read = '1') then                                           
        -- Das Hilfs-FPGA liest eine Speicherzelle aus und treibt dazu den Adressbus.
        -- Falls es aus dem Adressbereich von Display oder Tastatur liest, muss es den Datenwert selbst
        -- erzeugen (die ersten beiden Faelle), andernfalls tut dies SRAM, EPROM, parallele oder serielle I/O.
        -- Der Democom wird automatisch von den Bussen abgeklemmt.
        
          address <= areg_rollin(19 downto 0);                                
          
          if(chipsel(0) & address(1 downto 0) & m_oe & m_we = "00001") then   -- | Lesen einer Adresse auf "00" im
            databus <= dreg_rollin(15 downto 1) & user_disp_busy;             -- | Adressbereich des Displays 
                                     
          elsif((chipsel(5)='0') AND (m_oe='0') AND (m_we = '1')) then        -- Lesen aus dem Adressbereich der Tastatur
            if (address(0) = '0') then                       -- gerade Adresse -> tast_reg(15..0) ausgeben (dann loeschen)
              databus <= tast_reg(15 downto 0);
            else                                             -- ungerade Adresse -> tast_reg(19..16) ausg. (dann loeschen)
              databus <= dreg_rollin(15 downto 4) & tast_reg(19 downto 16);
            end if;
          else
            databus <= "ZZZZZZZZZZZZZZZZ";
          end if;
          
          
        else
        -- "Standardfall": Die Busse sind zur allgemeinen Benutzung freigegeben und die Adresse wird 
        -- vom Hilfs-FPGA nicht getrieben. Auf dem Datenbus werden nur Daten erzeugt, falls der Democom
        -- aus dem Adressbereich von Display oder Tastatur liest.
        
          address <= "ZZZZZZZZZZZZZZZZZZZZ";
          
          if(chipsel(0) & address(1 downto 0) & m_oe & m_we = "00001") then   -- | Lesen einer Adresse auf "00" im
            databus <= dreg_rollin(15 downto 1) & user_disp_busy;             -- | Adressbereich des Displays 
                                     
          elsif((chipsel(5)='0') AND (m_oe='0') AND (m_we = '1')) then        -- Lesen aus dem Adressbereich der Tastatur
            if (address(0) = '0') then                       -- gerade Adresse -> tast_reg(15..0) ausgeben (dann loeschen)
              databus <= tast_reg(15 downto 0);
            else                                             -- ungerade Adresse -> tast_reg(19..16) ausg. (dann loeschen)
              databus <= dreg_rollin(15 downto 4) & tast_reg(19 downto 16);
            end if;
          else
            databus <= "ZZZZZZZZZZZZZZZZ";
          end if;
          
        end if;
    end process;
    
    
    ADRESSREGISTER: process(clk_invers, reset, arollin, arollin_mux, a_or_d_changed, m_reg_ack)
    ---------------------------------------------------------------------------------------------------------
    -- Dieser Prozess beschreibt die beiden Register, welche Adressen puffern.
    --
    -- Zur Erinnerung: Adressbus (address)        --/20--->  areg_muxout  --/8--->  PC-Interface (pc_if_data)
    --                 PC-Interface (pc_if_data)  --/2*4-->  areg_rollin  --/20-->  Adressbus (address)
    ---------------------------------------------------------------------------------------------------------
    begin
        if (reset = '0') then
            areg_muxout <= "00000000000000000000";
            areg_rollin <= "000000000000000000000";
        elsif (clk_invers'event and clk_invers='1') then  -- Rueckflanke
            if (arollin = '1') then                       -- um 4 Bit nach links shiften
                areg_muxout <= areg_muxout;
                areg_rollin(20)           <= areg_rollin(16);
                areg_rollin(19 downto 16) <= areg_rollin(15 downto 12);
                areg_rollin(15 downto 12) <= areg_rollin(11 downto  8);
                areg_rollin(11 downto  8) <= areg_rollin( 7 downto  4);
                areg_rollin( 7 downto  4) <= areg_rollin( 3 downto  0);
                case arollin_mux is
                    when '1'    => areg_rollin( 3 downto  0) <= pc_if_data(7 downto 4);
                    when others => areg_rollin( 3 downto  0) <= pc_if_data(3 downto 0);
                end case;  
            elsif (a_or_d_changed = '1') then             -- gueltige Adresse oder Democom-Register liegen an
                areg_muxout <= address;
                areg_rollin <= areg_rollin;
            end if;
        end if;
    end process;

    DATENREGISTER: process(clk_invers, reset, drollin, drollin_mux)
    ---------------------------------------------------------------------------------------------------------
    -- Folgende beiden Prozesse beschreiben die beiden Register, welche Daten puffern.
    -- Zur Erinnerung: Datenbus (databus)         --/16--->  dreg_muxout  --/8--->  PC-Interface (pc_if_data)
    --                 PC-Interface (pc_if_data)  --/2*4-->  dreg_rollin  --/16-->  Datenbus (databus)
    ---------------------------------------------------------------------------------------------------------
    begin
        if (reset = '0') then
            dreg_rollin <= "0000000000000000";
        elsif (clk_invers'event and clk_invers='1') then  -- Rueckflanke
            if (drollin = '1') then                       -- 4 Bit nach links shiften
                dreg_rollin(15 downto 12) <= dreg_rollin(11 downto  8);
                dreg_rollin(11 downto  8) <= dreg_rollin( 7 downto  4);
                dreg_rollin( 7 downto  4) <= dreg_rollin( 3 downto  0);
                case drollin_mux is
                    when '1'    => dreg_rollin( 3 downto  0) <= pc_if_data(7 downto 4);
                    when others => dreg_rollin( 3 downto  0) <= pc_if_data(3 downto 0);
                end case;                
            end if;
        end if;
    end process;
    
    DATENREGISTER2: process(clk, reset, a_or_d_changed, m_oe, m_we, m_reg_ack)
    begin
        if (reset = '0') then
            dreg_muxout <= "0000000000000000";
        elsif (clk'event and clk='1') then                -- Vorderflanke
            if (a_or_d_changed = '1') then                -- gueltige Daten oder Democom-Register liegen an
                dreg_muxout <= databus;
            end if;
        end if;
    end process;
        
--- RAM-Zugriffsmaschine Anfang ---
    mem_ready <= '1' when ((c_rasta = read_ready) OR (c_rasta = write_ready)) else '0';
                                   -- Abschluss eines Lese-/Schreibvorganges signalisieren
   
    RAM_IF_STEUERSIGNALE: process (c_rasta, m_highz_t)
    begin
        if (m_highz_t = '1') then  -- Wenn externen Komponenten (Democom, SRAM etc.) _nicht_ befohlen wird, die Busse
                                   -- freizugeben, darf das Hilfs-FPGA diese (siehe Prozess BUSSE) und die Steuer-
                                   -- signale m_oe und m_we natuerlich nicht treiben.
          m_oe <= 'Z';
          m_we <= 'Z';
        else                       -- Externe geben Busse frei, Hilfs-FPGA steuert SRAM etc. an.
          case c_rasta is
            when read_init | read_ready  =>  m_oe <= '0'; m_we <= '1';                                                             
            when write_init              =>  m_oe <= '1'; m_we <= '0';
            when others                  =>  m_oe <= '1'; m_we <= '1';
         end case;
        end if;
    end process;
   
    ZUGRIFF: process(c_rasta, mem_read, mem_write)
    begin
        case c_rasta is
            when warten => if (mem_read = '1') then
                               m_rasta <= read_init;
                           elsif (mem_write = '1') then
                               m_rasta <= write_init;
                           else
                               m_rasta <= warten;
                           end if;
            when read_init   => m_rasta <= read_ready;
            when read_ready  => m_rasta <= warten;
            when write_init  => m_rasta <= write_ready;
            when write_ready => m_rasta <= warten;
            when others      => m_rasta <= warten;                -- | Zur Sicherheit: Falls der Automat den Bereich der
        end case;                                                 -- | erlaubten Zustandskodierungen verlassen sollte,
    end process;                                                  -- | wird er zurueckgesetzt.
   
    RAM_IF_SYNCH: process(clk_invers, reset)
    begin
        if (reset = '0') then
            c_rasta <= warten;
        elsif (clk_invers'event and clk_invers='1') then  -- Rueckflanke
            c_rasta <= m_rasta;
        end if;
    end process;
--- RAM-Zugriffsmaschine Ende ---
                
end behaviour;
