-- DEMOCOM/PRIMA '95
-- Interface PC <-> Hilfs-FPGA (8 Bit bidir., 4 Steuerleitungen)            C.K.
-- 
-- V0.1   17.04.95
-- V0.15  18.04.95  Zum Ausdrucken
-- V0.16  19.04.95  00.04 Uhr
-- V0.19  19.04.95  21.35 Uhr - morgen grosse Fehlersuche!
-- V0.2a  01.05.95  Hoffentlich alle Fertigkeiten abgedeckt, Beginn der Simulation
-- V0.22  05.05.95  Absoluter Wurstcode ist entstanden, da nun zwei teilparallele FSMs
-- V1.0   19.10.95  Final version (so wurstig nimmer)
-- V1.1   15.11.95  Alle Features funktionieren zuverlaessig (eingeschlossen Konfiguration, Tastatur, Display)
-- V1.11  24.11.95  Kleine Optimierungen
-- V1.3   22.01.96  Ueberarbeitete Kommentare

-- Das Hilfs-FPGA geht in den Modus "single step", als dessen Folge der Democom keine Clock mehr bekommt,
-- falls "single step" explizit angefordert wird oder falls ein Konfigurationsbefehl (Start, Daten, Ende) geschickt
-- wird; dadurch ist der Democom zunaechst einmal angehalten und kann mit "ss" (Kommando der PC-Software)
-- weiter taktweise beobachtet werden oder mit "si"/"halt" in den Halt-Modus gebracht werden.


-- Der Befehl "halt" der PC-Software bewirkt nun stets dasselbe wie "si": Die zweite FSM durchlaeuft zuerst
-- (im Prozess ACK_KONTROLLE) den Zustand "sinstr", wodurch das Status-Register auf "running" gesetzt wird
-- (d.h. der Democom bekommt eine normale Clock und das Signal m_halt wird deaktiviert). Dieses schadet nicht,
-- wenn der Democom sowieso schon lief, der Befehl also frueher "halt" geheissen haette.
-- War der Democom schon "halted", ist das Deaktivieren von m_halt notwendig, damit der naechste Democom-Befehl
-- begonnen wird (m_halt kann fuer einen/zwei Takte deaktiviert werden durch "halt"/"halt2", s.u.).
-- Danach nimmt die zweite FSM den Zustand "sinstr_wait" ein, in welchem das Status-Register auf "halted" gesetzt
-- und auf die Bestaetigung vom Democom gewartet wird, dass dieser tatsaechlich (wieder) angehalten hat.
-- Abschliessend koennen die zuletzt gepufferten Adressen und Daten an den PC gesendet werden.


-- Erweiterung von "halt"/"si": Ist das unterste Bit des Befehlswortes 0, wird m_halt fuer einen, ansonsten fuer
-- zwei Takte deaktiviert ('1'). Dies erlaubt es einem bestimmten studentischen Design, den Halte-Zustand
-- "weit genug" zu verlassen, um nicht gleich wieder in selbigem zu landen _ohne_ ueberhaupt einen Befehl
-- ausgefuehrt zu haben.


--------------------------------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity pc_if is
    port(
        reset:          in std_logic;                        -- KEIN Power-On-Reset, sondern zur Initialisierung 
                                                             -- nach der Konfiguration oder spaeter
                                                             
        clk:            in std_logic;                        -- globale Clock fuer alle speichernden Elemente
        clk_invers:     in std_logic;
        
        progr_m_status: in std_logic;                        -- Rueckmeldung bei Konfiguration des 81500:
                                                             -- kurz '0' zu Beginn; waehrend der Konf. '0', falls Fehler
                                                             
        progr_conf_done: in std_logic;                       -- Bestaetigung vom 81500: Konfiguration beendet
        
        m_obf:          in std_logic;                        -- "Output buffer full" am 8255, d.h. ein Befehl oder Daten
                                                             -- vom PC liegen im 8255 zur Abholung bereit
                                                             
        ibf:            in std_logic;                        -- "Input buffer full" am 8255, d.h. Hilfs-FPGA hat Daten
                                                             -- zum 8255 geschrieben, die noch nicht abgeholt worden sind.
                                                             
        mem_ready:      in std_logic;                        -- Bestaetigung von RAM-Zugriffsmaschine (in dieser Datei):
                                                             -- Schreib-/Lesevorgang abgeschlossen
        
        m_reg_ack:      in std_logic;                        -- Bestaetigung vom Democom:  gewuenschtes Register liegt an
        
        areg_muxout:    in std_logic_vector(19 downto 0);    -- Puffert bei (a_or_d_changed = '1') die Adresse
        dreg_muxout:    in std_logic_vector(15 downto 0);    -- Puffert bei (a_or_d_changed = '1') die Daten
        
        m_halt_ack:     in std_logic;                        -- Bestaetigung vom Democom: habe angehalten
        
        pc_if_data_in:  in std_logic_vector(7 downto 0);     -- Daten zwischen Hilfs-FPGA und PC - eingehende Richtung
        pc_if_data_out: out std_logic_vector(7 downto 0);    -- Daten zwischen Hilfs-FPGA und PC - ausgehende Richtung
        
        mem_read:       out std_logic;                       -- |Aufforderung an RAM-Zugriffsmaschine (in dieser Datei):
        mem_write:      out std_logic;                       -- |eine Speicherzelle beschreiben/auslesen
        
        m_ack:          out std_logic;                       -- "Acknowledge" am 8255, d.h. 8255 treibt Datenbus
        m_stb:          out std_logic;                       -- "Strobe" am 8255, d.h. Daten werden in 8255 uebernommen
        
        progr_m_config: out std_logic;                       -- Beginn der Konfiguration des Democom
        progr_dclk:     out std_logic;                       -- Clock fuer Konfiguration des Democom
        
        m_reg_req:      out std_logic;                       -- |Aufforderung an Democom: das durch reg_select
        reg_select:     out std_logic_vector(3 downto 0);    -- |bezeichnete Register auf Adress- ODER Datenbus legen
        
        arollin:        out std_logic;                       -- | 20 bzw. 16 Bit breite Register [a/d]reg_rollin 
        drollin:        out std_logic;                       -- | werden mit Daten vom PC-Interface bei ([a/d]rollin='1') 
        arollin_mux:    out std_logic;                       -- | und Rueckflanke vierbitweise gefuellt; [a/d]rollin_mux 
        drollin_mux:    out std_logic;                       -- | waehlt ein Nibble von pc_if_data aus.
        
        pc_if_driven:   out std_logic;                       -- Output Enable fuer pc_if_data_out
        
        clk_aktiv:      out std_logic;                       -- Democom-Clock <= (clk AND clk_aktiv);
        
        m_halt:         out std_logic;                       -- Aufforderung an Democom: vor naechstem Befehl anhalten
        
        progr_data:     out std_logic;                       -- 1 Bit Daten fuer (serielle) Konfiguration des Democom
         
        reset_democom:  out std_logic );                     -- Reset, der an den Democom geht - 
                                                             -- KEIN Power-On-Reset, sondern zur Initialisierung 
                                                             -- nach Einschalten und Konfiguration des Democom
end pc_if;


architecture behaviour of pc_if is

type state_type is (wait_for_event, fetch, decode, sstep, sstep_pulse, go, democom_reset,
                    sto_wait, sto_hi, sto_lo, write_mem, read_mem,
                    read_mem_hi_wait, read_mem_hi, read_mem_lo_wait, read_mem_lo, read_mem_lo_hold,
                    config, config2, config_wait1, config_wait2, config_waitf, config_fetch, config_rise, config_fall,
                    config_end, bericht, bericht_send, bericht_clear,
                    err_unknown, err_notnow,
                    ack_fsm_sinstr, ack_fsm_read_reg,
                    abort, areg_hi_wait, areg_hi, areg_mid_wait, areg_mid, areg_lo_wait, areg_lo);     -- 40 Zustaende

signal c_state, m_state, c_state_buffer: state_type;

type ack_state_type is (sinstr, sinstr2, sinstr_wait, read_reg, warten, senden, ende);                 -- 7 Zustaende

signal c_state2, m_state2, c_state2_buffer: ack_state_type;                        

signal i_reg: std_logic_vector(7 downto 0);  -- wird im Zustand fetch mit pc_if_data geladen

signal counter: natural range 0 to 8;            -- Ein Zaehler fuer drei Aufgaben: Bei der Konfiguration des Democom
                                                 -- muss zweimal mind. 5 us=Takte gewartet werden; beim Senden von
                                                 -- areg_rollin/dreg_rollin wird die Anzahl uebertragener Bytes
                                                 -- mitgezaehlt (3 bzw. 2); byteweise angelieferte Konfigurationsdaten
                                                 -- werden zu einzelnen Bits durch achtfaches Schieben von konf_reg.

signal status_reg: std_logic_vector(7 downto 0); -- Status-Register: 7    - Waiting for m_halt_ack from Democom
                                                 --                  6    - Waiting for m_reg_ack from Democom
                                                 --                  5..4 - Running(00)/Halted(10)/Single Step(11)
                                                 --                  3    - Unknown command encountered
                                                 --                  2    - Command illegal in this mode
                                                 --                  1..0 - Configuration successful (00)
                                                 --                         No Acknowledge (01; Startwert)
                                                 --                         Acknowledge too early (10)
                                                 --                         General Error (11; progr_m_status='0')
                                                 
signal konf_reg: std_logic_vector(7 downto 0);   -- Speichert 8 Bits fuer die Konfiguration (Programmierung) des
                                                 -- Democom-FPGAs; werden auf einmal vom pc_if gelesen, dann aber
                                                 -- seriell ausgegeben (konf_reg ist Schieberegister).  

signal sstep_pulse_ff: std_logic;                -- Das Signal clk_aktiv soll sich mit der Rueckflanke aendern, ist aber
                                                 -- abhaengig von (c_state = sstep_pulse). c_state jedoch aendert
                                                 -- sich mit der Vorderflanke, so dass (c_state = sstep_pulse) ueber
                                                 -- sstep_pulse_ff um einen halben Takt verzoegert werden muss. 

signal progr_dclk_ff: std_logic;                 -- Verzoegert progr_dclk um einen halben Takt.

begin


--- Diverse Logik Anfang ---

    reset_democom <= '0' when (c_state = democom_reset) else '1';              -- Hardware-Reset an Mikroprozessor
    
    m_halt <= '0' when (status_reg(5 downto 4) = "10") else '1';               -- Status-Reg. wurde durch Halt / 
                                                                               -- Single Inst. auf "halted" gesetzt, 
                                                                               -- nun soll der Democom anhalten.
                      
    --- Democom bekommt Clock, wenn clk_aktiv '1' ist ---
    clk_aktiv <= '1' when ((status_reg(5 downto 4) = "00") OR                  -- Democom bekommt Clock wenn "running" oder
                           (status_reg(5 downto 4) = "10") OR                  -- "halted" oder
                           (status_reg(5 downto 4) = "11" AND sstep_pulse_ff = '1'))  -- ein "single step" gemacht wird.
                           else '0';    
 
    SSTEP_PULSE_FLIPFLOP: process(reset, clk_invers, c_state)  -- Wird fuer die Bildung von clk_aktiv benoetigt.
    begin
      if(reset = '0') then
          sstep_pulse_ff <= '0';
      elsif(clk_invers'event and clk_invers='1') then  -- Rueckflanke
           if (c_state = sstep_pulse) then
              sstep_pulse_ff <= '1';                                           -- clk_democom fuer einen Takt aktiv 
          else
              sstep_pulse_ff <= '0';
          end if;
      end if;
    end process;
    
    --- Signale fuer die Programmierung des Prozessor-FPGAs ---
    progr_data     <= konf_reg(0);                                   -- Progr.-Daten; LSB zuerst senden
    progr_m_config <= '0' when (c_state = config_wait1) else '1';    -- Progr. beginnen (durch Config Start)
    progr_dclk     <= progr_dclk_ff;                                 -- Clock fuer die Programmierung des Prozessor-FPGAs

    PROGRAMMIER_DATENCLOCK: process(reset, clk_invers, c_state)      -- Steuert progr_dclk_ff
    begin
        if (reset = '0') then
            progr_dclk_ff <= '0';
        elsif (clk_invers'event and clk_invers='1') then  -- Rueckflanke
            if (c_state = config_rise) then                          -- Wird bei Befehl Config Data durchlaufen.
                progr_dclk_ff <= '1';                                
            else
                progr_dclk_ff <= '0';                                -- | Bei Config Data folgt auf config_rise der
                                                                     -- | Zustand config_fall; hier wird konf_reg
                                                                     -- | geshiftet -> neues Bit liegt an.
            end if;                 
        end if;
    end process;
    
    --- Aus dem 8255 lesen oder in ihn schreiben ---
    m_ack <= '0' when ((c_state = fetch) OR (c_state = sto_hi) OR    -- 8255 soll Bus treiben, Daten sind da (m_obf='0')
                       (c_state = sto_lo) OR (c_state = config_fetch)) 
                       else '1';                                     
                       
    m_stb <= '0' when ((c_state = areg_hi) OR (c_state = areg_mid) OR (c_state = areg_lo) OR  -- areg_muxout,
                       (c_state = read_mem_hi) OR (c_state = read_mem_lo) OR                  -- dreg_muxout bzw.
                       (c_state = bericht_send)) else '1';                                    -- status_reg in 8255
                                                                                              -- schreiben (pegelgest.)

    --- Schreiben/Lesen im Speicher ---
    mem_write <= '1' when (c_state = write_mem) else '0';
    mem_read  <= '1' when (c_state = read_mem)  else '0';

    --- Democom-Register auslesen ---
    m_reg_req <= '0' when (c_state2 = read_reg) else '1';
    reg_select <= i_reg(3 downto 0);
    
    REGISTER_SIGNALE: process(c_state, i_reg)
    begin
        if (i_reg(4) = '1') then                                    -- dreg_rollin beschreiben
            case c_state is
                when sto_hi => arollin <= '0'; arollin_mux <= '-';
                               drollin <= '1'; drollin_mux <= '1';  -- dreg_rollin(3 downto 0) <= pc_if_data(7 downto 4)
                               
                when sto_lo => arollin <= '0'; arollin_mux <= '-';
                               drollin <= '1'; drollin_mux <= '0';  -- dreg_rollin(3 downto 0) <= pc_if_data(3 downto 0)
                               
                when others => arollin <= '0'; arollin_mux <= '-';
                               drollin <= '0'; drollin_mux <= '-';
            end case;
        else                                                        -- areg_rollin beschreiben
            case c_state is
                when sto_hi => arollin <= '1'; arollin_mux <= '1';  -- areg_rollin(3 downto 0) <= pc_if_data(7 downto 4)
                               drollin <= '0'; drollin_mux <= '-'; 
                               
                when sto_lo => arollin <= '1'; arollin_mux <= '0';  -- areg_rollin(3 downto 0) <= pc_if_data(3 downto 0)
                               drollin <= '0'; drollin_mux <= '-';
                                                             
                when others => arollin <= '0'; arollin_mux <= '-';
                               drollin <= '0'; drollin_mux <= '-';
            end case;
        end if;
    end process;

--- Diverse Logik Ende ---

    ZUSTAENDE_PUFFERN: process(clk_invers, reset, c_state, c_state2, c_state_buffer, c_state2_buffer)
    begin
      if(reset = '0') then
        c_state_buffer  <= wait_for_event;
        c_state2_buffer <= warten;
      elsif(clk_invers'event and clk_invers='1') then             -- Rueckflanke
        c_state_buffer  <= c_state;
        c_state2_buffer <= c_state2;
      end if;
    end process;

    IREG: process(clk_invers, reset, c_state)  -- Instruktionsregister - aktuellen Befehl vom PC puffern
    begin
      if (reset = '0') then
          i_reg <= "00000000";
      elsif (clk_invers'event and clk_invers='1') then  -- Rueckflanke
        if (c_state = fetch) then
          i_reg <= pc_if_data_in;
        else
          i_reg <= i_reg;
        end if;
      end if;
    end process;
    
    KONFIGURATIONS_REGISTER: process(clk_invers, reset, c_state)  -- Zur Serialisierung der Konfigurationsdaten
    begin
        if (reset = '0') then
            konf_reg <= "00000000";
        elsif (clk_invers'event and clk_invers='1') then  -- Rueckflanke
            if (c_state = config_fetch) then
                konf_reg <= pc_if_data_in;
            elsif (c_state = config_fall) then
                konf_reg(0) <= konf_reg(1);
                konf_reg(1) <= konf_reg(2);
                konf_reg(2) <= konf_reg(3);
                konf_reg(3) <= konf_reg(4);
                konf_reg(4) <= konf_reg(5);
                konf_reg(5) <= konf_reg(6);
                konf_reg(6) <= konf_reg(7);
            else
                konf_reg <= konf_reg;
            end if;
        end if;
    end process;
    
    STATUS_REGISTER: process(clk_invers, reset, c_state, c_state2, progr_m_status, progr_conf_done)
    begin
        if (reset = '0') then
            status_reg <= "00000001";
        elsif (clk_invers'event and clk_invers='1') then  -- Rueckflanke
        
        ---- Wartet das Hilfs-FPGA auf den Democom? ----
            if (c_state = abort) then
              status_reg(7 downto 6) <= "00";                                    -- gegenwaertiges Warten abgebrochen
            else
              case c_state2 is
                when sinstr_wait => status_reg(7 downto 6) <= "10";              -- Warten auf m_halt_ack
                when read_reg    => status_reg(7 downto 6) <= "01";              -- Warten auf m_reg_ack 
                when senden      => status_reg(7 downto 6) <= "00";              -- Hilfs-FPGA wartet nicht (mehr)
                when others      => status_reg(7 downto 6) <= status_reg(7 downto 6);
              end case;
            end if; 
            
        ---- Laufmodus ----
            if ((c_state = go) OR (c_state2 = sinstr)) then
                status_reg(5 downto 4) <= "00";  -- "running"
            elsif (c_state2 = sinstr_wait) then
                status_reg(5 downto 4) <= "10";  -- "halted"
            elsif ((c_state = sstep) OR (c_state = config)) then
                status_reg(5 downto 4) <= "11";  -- | "single step"; wenn explizit angefordert ODER wenn Konfigurations-
                                                 -- | befehle gesendet werden, damit der Democom nach beendeter Konf.
                                                 -- | nicht gleich loslaeuft und noch ein Reset-Signal bekommen kann.
            else
                status_reg(5 downto 4) <= status_reg(5 downto 4);
            end if;
            
        ---- Fehlerbits ----    
            case c_state is
                when bericht_clear=> status_reg(3) <= '0';                  -- Fehlerbits loeschen
                                     status_reg(2) <= '0';
            
                when err_unknown  => status_reg(3) <= '1';                  -- Unbekannter Befehl
                                     status_reg(2) <= status_reg(2);
            
                when err_notnow   => status_reg(3) <= status_reg(3);
                                     status_reg(2) <= '1';                  -- Befehl in diesem Modus nicht zulaessig
            
                when others       => status_reg(3) <= status_reg(3);
                                     status_reg(2) <= status_reg(2);
            end case;

        ---- Zaehler fuer Konfigurationsergebnis (mit "merkwuerdigem Inkrementieren") ----
            ---------------------------------------------------------------------------------------------
            -- Grundidee: Es soll ueberprueft werden, ob genau die richtige, weil erwartete,
            -- Anzahl Konfigurationsbits gesendet worden ist.
            -- Abgefragt wird dazu progr_conf_done, welches nach dem letzten Bit dauerhaft '1' wird.
            -- 
            -- Vorgehen: Der Befehl Config Start macht den Zaehler mit der Ausgangsannahme
            -- "Signal progr_conf_done kam noch nicht" (Kode "01") scharf. prog_conf_done sollte nach Ein-
            -- takten des letzten Bits durch das letzte Config Data aktiv werden, was nicht mehr detektiert
            -- wird (Zustand config_fetch wird vorher erreicht); es folgt aber (ausreichend spaeter)
            -- noch ein Config End. Erreicht dieses config_end, wird Zaehler also genau einmal
            -- zum Endzustand von "Konfiguration war erfolgreich" (Kode "00") erhoeht.
            -- Schiefgehen koennen zwei Dinge: Hat das FPGA schon vorzeitig genug, wird progr_conf_done
            -- schon bei einem der Config Data-Befehle '1' sein und der Zaehler dauerhaft auf
            -- "Signal progr_conf_done kam zu frueh" (Kode "10") gesetzt. Einen "allgemeinen" Fehler kann
            -- das FPGA ueber die Leitung progr_m_status signalisieren, was den Zaehler ebenfalls 
            -- dauerhaft setzt, aber auf "Allgemeiner Fehler" (Kode "11").
           
            if ((c_state = config_fetch) OR (c_state = config_end)) then
            
                if (progr_m_status = '0') then                              -- | genereller Fehler -> FPGA zieht
                    status_reg(1 downto 0) <= "11";                         -- | progr_m_status dauerhaft auf '0'
                                                                            
                elsif (progr_conf_done = '1') then                          -- Merkwuerdiges Inkrementieren
                
                    if (status_reg(1 downto 0) = "01") then        -- "Signal progr_conf_done kam noch nicht"
                        status_reg(1 downto 0) <= "00";
                    elsif (status_reg(1 downto 0) = "00") then     -- | "Konfiguration war erfolgreich"; erst beim
                                                                   -- | Ausfuehren von Config End war progr_conf_done '1'
                        status_reg(1 downto 0) <= "10";
                    else
                        status_reg(1 downto 0) <= status_reg(1 downto 0);
                    end if;
                end if;
                
            elsif (c_state = config_wait1) then                             -- | Zuruecksetzen durch Config Start auf
                status_reg(1 downto 0) <= "01";                             -- | Ausgangsannahme "Signal
                                                                            -- | progr_conf_done kam noch nicht"
            else
                status_reg(1 downto 0) <= status_reg(1 downto 0);
            end if;
        end if;
    end process;
        
    INTERFACE_BUS: process(c_state, status_reg, areg_muxout, dreg_muxout)
    begin
        case c_state is
        
            -- Die zwei Zustaende in jedem Zweig der case-Anweisung folgen im ersten Automaten direkt
            -- aufeinander. Im jeweils ersten Zustand ist (m_stb = '0'), so dass der 8255 den entsprechenden Wert 
            -- latcht, der zweite ist noetig, um die Haltezeit von pc_if_data_out nach aktivem m_stb einzuhalten.
            
            when read_mem_hi | read_mem_lo_wait => pc_if_data_out <= dreg_muxout(15 downto  8);
                                                   pc_if_driven <= '1';  -- Bus zum PC wird mit diesem Signal getrieben
                                                   
            when read_mem_lo | read_mem_lo_hold => pc_if_data_out <= dreg_muxout( 7 downto  0);
                                                   pc_if_driven <= '1';  --                - " -
                                                   
            when areg_hi     | areg_mid_wait    => pc_if_data_out <= "0000"&areg_muxout(19 downto 16);
                                                   pc_if_driven <= '1';  --                - " -
                                                   
            when areg_mid    | areg_lo_wait     => pc_if_data_out <= areg_muxout(15 downto  8);
                                                   pc_if_driven <= '1';  --                - " -
                                                   
            when areg_lo     | read_mem_hi_wait => pc_if_data_out <= areg_muxout( 7 downto  0);
                                                   pc_if_driven <= '1';  --                - " -
                                                   
            when bericht_send | bericht_clear   => pc_if_data_out <= status_reg;
                                                   pc_if_driven <= '1';  --                - " -
                                                   
            when others       => pc_if_data_out <= dreg_muxout(15 downto 8); -- nur ein Default-Wert
                                                   pc_if_driven <= '0';      -- in ctrl_pur: pc_if_data_out <= "ZZZZZZZZ"
        end case;
    end process;
    
    ZAEHLER: process(clk_invers, reset, c_state, counter)  -- Multifunktionszaehler, s. Signaldeklaration
    begin
        -- Zuruecksetzen des Zaehlers erfolgt implizit dadurch, dass der erste Automat nach Reset
        -- im Zustand wait_for_event beginnt.
        
        if (clk_invers'event and clk_invers = '1') then  -- Rueckflanke
          case c_state is
                when config_wait1 | config_wait2 | config_rise | sto_hi
                                                                => if (counter = 8) then
                                                                       counter <= 8;
                                                                   else
                                                                       counter <= counter + 1;
                                                                   end if;
                                                                   
                when config2 | config_fetch | wait_for_event    => counter <= 0;
                
                when others                                     => counter <= counter;
            end case;
        end if;
    end process;
    
    KONTROLLE: process(c_state, c_state2_buffer, m_obf, ibf, i_reg, mem_ready, counter, status_reg)
    -- Dies ist der erste Automat.
    -- Er holt einen neuen Befehl vom PC-Interface ab, dekodiert diesen und fuehrt ihn aus.
    begin
      case c_state is
        when wait_for_event   => if(c_state2_buffer = senden) then
                                   m_state <= areg_hi_wait;       -- 3+2 Bytes senden, OBF nicht leeren
                                 elsif (m_obf = '0') then
                                   m_state <= fetch;              -- PC hat Befehl zum Port gesandt!
                                 else
                                   m_state <= wait_for_event;     -- Nix passiert...
                                 end if;
                                                                  -- Vierter Fall: Breakpoint-Bedingung erfuellt
                                                                  -- !!! (Aktion bei IRQs?) - noch nicht implementiert
        when fetch            => m_state <= decode;
        when decode         => if(status_reg(7 downto 6) = "00") then  -- Hilfs-FPGA wartet _nicht_ auf Democom
                                 case status_reg(5 downto 4) is
                                   when "00" =>                                 -- Democom laeuft - "running"
                                     case i_reg(7 downto 4) is
                                       when "0000" => m_state <= ack_fsm_sinstr;     -- Anhalten -> "halted"
                                       when "0001" => m_state <= err_notnow;         -- Single Step - erst anhalten!
                                       when "0101" => m_state <= ack_fsm_sinstr;     -- Anhalten -> "halted"
                                       when "0111" => m_state <= go;                 -- hier redundant; meinetwegen...
                                       when "0010"|"0011" => m_state <= wait_for_event;  -- Tastenabfrage an/aus - ignor.
                                       when "1000"|"1001" => m_state <= sto_wait;          -- areg-/dreg_rollin setzen
                                       when "1010" => m_state <= err_notnow;         -- write_mem - erst anhalten!
                                       when "1011" => m_state <= err_notnow;         -- read_mem  -  "      "
                                       when "0100" => m_state <= config;             -- Brutal, aber was soll's
                                       when "1100" => m_state <= bericht;            -- Status-Byte an PC senden
                                       when "1101" => m_state <= err_notnow;         -- read_reg - anh., Daten ueber databus!
                                       -- when "111-" => m_state <= set_bp;          -- Breakpoint setzen (Adr./Datum)
                                       --  mit "0110" meldet das Hilfs-FPGA das Auftreten einer Breakpointbedingung
                                       when "1110" => m_state <= democom_reset;      -- low-aktiven Reset an Democom senden
                                       when "0110" => m_state <= abort;              -- Die Ack-FSM und diese zuruecksetzen
                                       when "1111" => m_state <= err_unknown;
                                       when others => m_state <= err_unknown;        -- Bei unbek. Befehl Status-Bit setzen                                     
                                     end case;
                                   when "10" =>                                 -- Democom angehalten - "halted"
                                     case i_reg(7 downto 4) is
                                       when "0000" => m_state <= ack_fsm_sinstr;     -- Genau eine Instruktion ausfuehren
                                       when "0001" => m_state <= sstep;              -- Genau eine Clock -> "single step"
                                       when "0101" => m_state <= ack_fsm_sinstr;     -- Genau eine Instruktion ausfuehren
                                       when "0111" => m_state <= go;                 -- Wieder normal laufen -> "running"
                                       when "0010"|"0011" => m_state <= wait_for_event;  -- Tastenabfr. an/aus - ignor.
                                       when "1000"|"1001" => m_state <= sto_wait;    -- areg-/dreg_rollin setzen
                                       when "1010" => m_state <= write_mem;
                                       when "1011" => m_state <= read_mem;
                                       when "0100" => m_state <= config;
                                       when "1100" => m_state <= bericht;            -- Status-Byte an PC senden
                                       when "1101" => m_state <= ack_fsm_read_reg;   -- Democom-Reg. => dreg_muxout => PC
                                       when "1110" => m_state <= democom_reset;      -- low-aktiven Reset an Democom senden
                                       when "0110" => m_state <= abort;              -- Die Ack-FSM und diese zuruecksetzen
                                       when "1111" => m_state <= err_unknown;
                                       when others => m_state <= err_unknown;        -- Bei unbek. Befehl Status-Bit setzen
                                     end case;
                                   when "11" =>                                 -- Democom im "Single Step"-Modus
                                     case i_reg(7 downto 4) is
                                       when "0000" => m_state <= ack_fsm_sinstr;     -- Instruktion beenden -> "halted"
                                       when "0001" => m_state <= sstep;              -- Genau eine Clockflanke geben
                                       when "0101" => m_state <= ack_fsm_sinstr;     -- Instruktion beenden -> "halted"
                                       when "0111" => m_state <= go;                 -- Wieder normal laufen -> "running"
                                       when "0010" | "0011" => m_state <= wait_for_event;  -- Tastenabfr. an/aus - ignor.
                                       when "1000" | "1001" => m_state <= sto_wait;   -- areg-/dreg_rollin setzen
                                       when "1010" => m_state <= err_notnow;         -- Kein Rumgefummel bei Single Step!
                                       when "1011" => m_state <= err_notnow;         -- Kein Rumgefummel bei Single Step!
                                       when "0100" => m_state <= config;             -- | Wichtig, dass auch hier erlaubt,
                                                                                     -- | denn status_reg wird durch  
                                                                                     -- | config auf "Single Step" gesetzt!
                                       when "1100" => m_state <= bericht;            -- Status-Byte an PC senden
                                       when "1101" => m_state <= err_notnow;         -- Kein Rumgefummel bei Single Step!
                                       when "1110" => m_state <= democom_reset;      -- low-aktiven Reset an Democom senden
                                       when "0110" => m_state <= abort;              -- Die Ack-FSM und diese zuruecksetzen
                                       when "1111" => m_state <= err_unknown;
                                       when others => m_state <= err_unknown;        -- Bei unbek. Befehl Status-Bit setzen
                                     end case;
                                   when "01"   => m_state <= err_unknown;            -- Falls im Status-Reg. Unsinn steht...
                                   when others => m_state <= err_unknown;
                                 end case;
                               else                                             -- Paralleler Befehl laueft in Ack-FSM
                                 case i_reg(7 downto 4) is
                                   when "1100" => m_state <= bericht;                -- Status-Byte an PC senden             
                                   when "0110" => m_state <= abort;                  -- beide Automaten zuruecksetzen
                                   when others => m_state <= err_notnow;
                                 end case;
                               end if;
                               
        when err_unknown      => m_state <= wait_for_event;         -- Status-Bit setzen, Befehl ignorieren
        when err_notnow       => m_state <= wait_for_event;         --                - " -

        when sstep            => m_state <= sstep_pulse;            -- status_reg auf "Single Step"
        when sstep_pulse      => m_state <= areg_hi_wait;           -- Clock-Impuls erzeugen, dann Adr. & Daten ausgeben

        when go               => m_state <= wait_for_event;
        when democom_reset    => m_state <= wait_for_event;
        
        when sto_wait        => if (m_obf = '0') then               -- areg_rollin/dreg_rollin vom PC lesen
                                  m_state <= sto_hi;
                                else
                                  m_state <= sto_wait;
                                end if;
        when sto_hi           => m_state <= sto_lo;                 -- counter ++ 1;
        when sto_lo           => if (i_reg(4) = '1') then           -- Daten (16 Bit) einlesen
                                   if (counter >= 2) then
                                     m_state <= wait_for_event;
                                   else
                                     m_state <= sto_wait;
                                   end if;
                                 else                               -- Adresse (20+4 Bit) einlesen
                                   if (counter >= 3) then
                                     m_state <= wait_for_event;
                                   else
                                     m_state <= sto_wait;
                                   end if;
                                 end if;
                                 
        when write_mem        => if (mem_ready = '1') then
                                   m_state <= wait_for_event;
                                 else
                                   m_state <= write_mem;
                                 end if;
                                 
        when read_mem         => if (mem_ready = '1') then
                                   m_state <= read_mem_hi_wait;     -- 16 Bit Daten liegen an, nun ins Reg. laden
                                 else
                                   m_state <= read_mem;             -- Speicher troedelt noch
                                 end if;
        when read_mem_hi_wait => if (ibf = '0') then                -- pc_if_data_out <= areg_muxout(7 downto 0); wg Hold
                                   m_state <= read_mem_hi;
                                 else
                                   m_state <= read_mem_hi_wait;
                                 end if;
        when read_mem_hi      => m_state <= read_mem_lo_wait;       -- pc_if_data_out <= dreg_muxout(15 downto 8);
        when read_mem_lo_wait => if (ibf = '0') then                -- pc_if_data_out <= dreg_muxout(15 downto 8); wg Hold
                                   m_state <= read_mem_lo;
                                 else
                                   m_state <= read_mem_lo_wait;
                                 end if;
        when read_mem_lo      => m_state <= read_mem_lo_hold;       -- pc_if_data_out <= dreg_muxout(7 downto 0);
        when read_mem_lo_hold => m_state <= wait_for_event;         -- pc_if_data_out <= dreg_muxout(7 downto 0); wg Hold

        when config           => case i_reg(3 downto 1) is          -- counter <= 0;
                                   when "001" => m_state <= config_wait1; -- 81500 in Prog.-Modus bringen via prog_m_conf
                                   when "010" => m_state <= config_waitf; -- ein Datenbyte PC -> Hilfs-FPGA -> 81500
                                   when "100" => m_state <= config_end;  
                                   when others => m_state <= err_unknown;
                                 end case;         
        when config_wait1     => if (counter = 8) then              -- prog_m_config <= '0'; counter ++ 1; mind. 5 us warten
                                   m_state <= config2;
                                 else
                                   m_state <= config_wait1;
                                 end if;
        when config2          => m_state <= config_wait2;           -- counter <= 0;
        when config_wait2     => if (counter = 8) then              -- prog_m_config <= '1'; counter ++ 1; mind. 5 us warten
                                   m_state <= wait_for_event;
                                 else
                                   m_state <= config_wait2;
                                 end if;
        
        when config_waitf     => if (m_obf = '0') then              -- 81500 ist im Config-Modus; Warten auf naechstes Wort
                                   m_state <= config_fetch;
                                 else
                                   m_state <= config_waitf;
                                 end if;
        when config_fetch     => m_state <= config_rise;            -- counter <= '0';
        when config_rise      => m_state <= config_fall;            -- progr_dclk <= '1'; konf_reg[0]->81500; counter ++ 1;
        when config_fall      => if (counter = 8) then              -- progr_dclk <= '0'; konf_reg shiften
                                   m_state <= wait_for_event;
                                 else
                                   m_state <= config_rise;
                                 end if;
        
        when config_end       => m_state <= bericht;                -- conf_m_done kontrollieren, dann status_reg senden
        
        when bericht          => if (ibf = '0') then
                                   m_state <= bericht_send;
                                 else
                                   m_state <= bericht;
                                 end if;
        when bericht_send     => m_state <= bericht_clear;          -- pc_if_data_out <= "00"&status_reg;
        when bericht_clear    => m_state <= wait_for_event;         -- Fehlerbits im Status-Reg. loeschen
        
        when ack_fsm_sinstr   => m_state <= wait_for_event;         -- | Zweiten Automaten mit Warten auf m_halt_ack 
        when ack_fsm_read_reg => m_state <= wait_for_event;         -- | bzw. m_reg_ack beauftragen.
        
        when abort            => m_state <= wait_for_event;         -- Auftrag des zweiten Automaten abbrechen.
        
        when areg_hi_wait     => if (ibf = '0') then
                                   m_state <= areg_hi;
                                 else
                                   m_state <= areg_hi_wait;
                                 end if;
        when areg_hi          => m_state <= areg_mid_wait;          -- pc_if_data_out <= "0000"&areg_muxout(19 downto 16);
        when areg_mid_wait    => if (ibf = '0') then                --                   --- " ---; wg Hold
                                   m_state <= areg_mid;
                                 else
                                   m_state <= areg_mid_wait;
                                 end if;
        when areg_mid         => m_state <= areg_lo_wait;           -- pc_if_data_out <= areg_muxout(15 downto 8);
        when areg_lo_wait     => if (ibf = '0') then                --                   --- " ---; wg Hold
                                   m_state <= areg_lo;
                                 else
                                   m_state <= areg_lo_wait;
                                 end if;
        when areg_lo          => m_state <= read_mem_hi_wait;       -- pc_if_data_out <= areg_muxout(7 downto 0);
        
        when others           => m_state <= wait_for_event;         -- | Zur Sicherheit: Falls der Automat den Bereich der
                                                                    -- | erlaubten Zustandskodierungen verlassen sollte,
                                                                    -- | wird er zurueckgesetzt.

      end case;
    end process;

    ACK_KONTROLLE: process(c_state2, i_reg, m_reg_ack, m_halt_ack, c_state_buffer)
    -- Dies ist der zweite Automat.
    -- Er kann konkurrent zum ersten auf die Handshaking-Signale m_halt_ack und m_reg_ack warten.
    begin
      case c_state2 is
        when warten      => case c_state_buffer is
                              when ack_fsm_sinstr   => m_state2 <= sinstr;
                              when ack_fsm_read_reg => m_state2 <= read_reg;
                              when others           => m_state2 <= warten;
                            end case;
        
        when sinstr      => if (i_reg(0) = '0') then                 -- | status_reg auf "running" => 
                                                                     -- | Democom-Clock an, m_halt <= '1';
                              m_state2 <= sinstr_wait;          
                            else
                              m_state2 <= sinstr2;                   -- Befehl "halt2" wurde abgesetzt.
                            end if;
                                 
        when sinstr2     => m_state2 <= sinstr_wait;                 -- m_halt bleibt noch einen Takt laenger auf '1'.
                                 
        when sinstr_wait => if (m_halt_ack = '0') then               -- m_halt <= '0'; auf Anhalten des Democom warten
                              m_state2 <= senden;                    -- status_reg auf "halted"
                            else
                              m_state2 <= sinstr_wait;
                            end if;
        
        when read_reg    => if (m_reg_ack = '0') then                -- auf Democom-Register warten
                              m_state2 <= senden;                    
                            else
                              m_state2 <= read_reg;                  -- Democom-Register troedelt noch
                            end if;
        
        when senden      => if (c_state_buffer = read_mem_lo) then   -- | Nach jedem der zwei von dieser FSM ausgefuehrten
                              m_state2 <= ende;                      -- | Befehle - Halt/SInstr und ReadReg - muessen 3+2
                            else                                     -- | Bytes gesendet werden (Adresse+Daten).
                              m_state2 <= senden;                    -- | Dies geschieht im ersten Automaten und ist 
                            end if;                                  -- | beendet im Zustand read_mem_lo(_hold).

        when ende        => m_state2 <= warten;
        
        when others      => m_state2 <= warten;                      -- | Zur Sicherheit: Falls der Automat den Bereich der
                                                                     -- | erlaubten Zustandskodierungen verlassen sollte,
                                                                     -- | wird er zurueckgesetzt.
      end case;
    end process;             
    
    SYNCH: process(clk, reset, m_state, m_state2)
    
    begin
        if (reset = '0') then
          c_state  <= wait_for_event;
          c_state2 <= warten;
        elsif (clk'event and clk='1') then          -- Vorderflanke
        
          if (m_state = abort) then                                 -- vergebliches Warten beenden durch "abort"
            c_state  <= m_state;
            c_state2 <= ende;
          else
            c_state  <= m_state;                                    -- normales Fortschreiten der Automaten
            c_state2 <= m_state2;
          end if;
          
        end if;
    end process;

end behaviour;
